// Author 		: F. de la Parra
// Date			: April, 2019
// Description  : Class ELinkedList implements a "DNode" doubly linked list. 
//				  DNode is an inner private class encapsulated in ELinkedList.
//				  It supports DNode insertions and deletions at both ends of 
//				  list, as well additional methods to find the list's size,
//				  traverse the list and display the contents of its nodes,
//				  determine whether a DNode of a given contents is in the list,
//				  determine whether the list is empty, and clear its content.
// Version      : 1.0


public class ELinkedList {
	
	// Inner class DNode. Template to instantiate the data nodes of the double-linke linked list.
	private class DNode {
		// Contents of data node.
		private String item;
		
		// Links from "this" DNode to previous and next Dnode. 
		private DNode prevLink;
		private DNode nextLink;
		
		// Default constructor for DNode inner class.
		public DNode() {
			prevLink = null;
			nextLink = null;
			item = null;
		}
		
		// Full parameter constructor for DNode inner class.
		public DNode(String newItem, DNode prevLinkValue, DNode nextLinkValue) {
			item = newItem;
			prevLink = prevLinkValue;
			nextLink = nextLinkValue;
		}
	}
	
	// Attributes of the ELinkedList outer class
	private DNode head;
	private DNode tail;
	
	// Default constructor for ELinkedList outer class.
	public ELinkedList() {
		head = null;		// Reference to start node of the list.
		tail = null;		// Reference to end node of the list.
	}
		
	// BEHAVIOUR METHODS
	
	// Add a DNode at the start of the linked list.
	public void addToStart(String itemName) {
		head = new DNode(itemName, null, head);
		if (tail == null)
			tail = head;
	}
	
	// Delete the head node.
	public boolean deleteHeadNode() {
		if (head != null) {
			if (head.nextLink != null) {
				head = head.nextLink;
				head.prevLink = null;
			}
			else {
				head = null;
				tail = null;
			}
			return true;
		}
		else
			return false;
		}
		
	// Add a DNode to the end of the linked list.
	public void addToEnd(String itemName) {
		DNode tempLink = new DNode(itemName, tail, null);
		if (tail != null) {
			tail.nextLink = tempLink;
			tail = tempLink;
		}
		else {
			head = tempLink;
			tail = tempLink;
			head.prevLink = null;
		}

	}
	
	// Delete the tail node.
	public boolean deleteTailNode() {
		if (tail != null) {
			if (tail.prevLink != null) { 
				tail = tail.prevLink;
				tail.nextLink = null;
			}
			else {
				head = null;
				tail = null;
			}
			return true;
		}
		else
			return false;	
	}
	
	// Compute the number of DNodes in the list.
	public int size(){
		int count = 0;
		DNode position = head;
		
		while (position != null) {
			count++;
			position = position.nextLink;
		}
		return count;
	}
	
	// Find out whether a DNode with a given content is in the linked list.
	public boolean contains(String item) {
		return (find(item) != null);
	}
 
	//Find the first DNode whose content is equal to a given target string.
	private DNode find(String target) {
		
		// Start at the entry point of the linked list.
		DNode position = head;
		String itemAtPosition;
		
		// Scan the linked list until the target is found.
		while (position != null) {
			itemAtPosition = position.item;
			if (itemAtPosition.equals(target))
				return position;
			position = position.nextLink;
		}
		
		return null; //target was not found
	}
	
	// List the contents of all the DNodes in the linked list.
	public void outputList() {
		DNode position = head;
		while (position != null) {
			System.out.println(position.item);
			position = position.nextLink;
		}
	}
	
	// Check whether the linked list is empty or not.
	public boolean isEmpty() {
		return (head == null);
	}
	
	// Clear the linked list.
	public void clear() {
		head = null;
		tail = null;
	}
}